extern crate xcb;
extern crate mpd;
extern crate simple_signal;

extern crate freetype;
extern crate fontconfig_sys;

mod sensors;
mod config;
mod ui;
mod widgets;

use simple_signal::Signal;

use std::env;
use std::process;
use std::sync::mpsc;
use std::thread;
use std::time;

use ui::x11;
use widgets::Message;

fn main() {
    let config_path = env::args().nth(1).unwrap_or("./panel.toml".to_string());
    let cfg = config::load(&config_path);

    if let Some(conn) = x11::Connection::new() {
        let (tx, rx) = mpsc::channel();
        {
            let tx = tx.clone();
            simple_signal::set_handler(&[Signal::Int, Signal::Term], move |_signals| {
                tx.send(Message::Quit).expect("Failed to send quit");
            });
        }
        {
            let tx = tx.clone();
            let conn = conn.clone();
            thread::spawn(move || {
                loop {
                    match conn.wait_for_event() {
                        Some(event) => {
                            let message = Message::XcbEvent(event);
                            tx.send(message).expect("Failed to send xcb event");
                        },
                        None => { break; }
                    }
                }
            });
        }

        let mut panel = ui::panel::Panel::new(conn, &cfg);
        let params = widgets::WidgetParams {
            id: "".to_string(),
            context: panel.make_context(),
            tx: tx.clone(),
            config: cfg.clone()
        };

        let left_widgets: Vec<widgets::WidgetConstructor> = vec![
            widgets::bspwm,
            widgets::spacer,
            widgets::title
        ];

        let right_widgets: Vec<widgets::WidgetConstructor> = vec![
            widgets::mpd,
            widgets::sensors,
            widgets::tray
        ];

        for cons in left_widgets {
            let widget = cons(params.clone());
            panel.add_left_widget(widget);
        }
        for cons in right_widgets {
            let widget = cons(params.clone());
            panel.add_right_widget(widget);
        }

        panel.create();

        {
            let tx = tx.clone();
            thread::spawn(move || {
                loop {
                    tx.send(Message::Update).expect("Failed to send update");
                    thread::sleep(time::Duration::from_secs(1));
                }
            });
        }

        loop {
            let event = rx.recv();
            if panel.handle_event(event.unwrap()) {
                println!("Exiting");
                break;
            }
        }
    }

    process::exit(0);
}
