use std::rc::Rc;

use ui::x11;
use ui::font;

use xcb;

#[derive(Clone)]
pub struct Context {
    pub conn: x11::Connection,
    pub height: u16,
    pub window: xcb::Window,
    picture: xcb::render::Picture,
    pen: xcb::render::Picture,
    fonts: Rc<font::FontLoader>,
    bg_color: xcb::render::Color,
    pub default_bg_color: xcb::render::Color
}

impl Context {
    pub fn new(
        conn: x11::Connection,
        height: u16,
        window: xcb::Window,
        picture: xcb::render::Picture,
        pen_color: xcb::render::Color,
        bg_color: xcb::render::Color,
        fonts: Rc<font::FontLoader>
    ) -> Self {
        let pen = conn.generate_id();
        xcb::render::create_solid_fill(
            &conn,
            pen,
            pen_color
        );

        Context {
            conn: conn,
            height: height,
            window: window,
            picture: picture,
            pen: pen,
            bg_color: bg_color,
            default_bg_color: bg_color,
            fonts: fonts
        }
    }

    pub fn draw_bg(&self, x: u16, width: u16) {
        xcb::render::fill_rectangles(
            &self.conn,
            xcb::render::PICT_OP_SRC as u8,
            self.picture,
            self.bg_color,
            &[xcb::Rectangle::new(x as i16, 0, width, self.height)]
        );
    }

    pub fn measure_text(&self, name: &str) -> u16 {
        let text = self.fonts.create_renderable_text(name);
        text.width
    }

    pub fn set_bg_color(&mut self, color: xcb::render::Color) {
        self.bg_color = color;
    }

    pub fn draw_text(&self, name: &str, x: u16) {
        if !name.is_empty() {
            let text = self.fonts.create_renderable_text(name);
            let baseline = self.height as i16 - self.fonts.default_offset(self.height);
            text.render(&self.conn, self.pen, self.picture, x, baseline as u16);
        }
    }

    pub fn draw_text_with_clipping(&self, name: &str, x: u16, width: u16) {
        xcb::render::set_picture_clip_rectangles(
            &self.conn,
            self.picture,
            0, 0,
            &[xcb::Rectangle::new(x as i16, 0, width, self.height)]
        );

        self.draw_text(name, x);

        xcb::render::change_picture(
            &self.conn,
            self.picture,
            &[(xcb::render::CP_CLIP_MASK, xcb::render::PICTURE_NONE)]
        );
    }
}
