use xcb;

use config::Config;
use ui::context::Context;
use ui::window::Window;
use ui::x11;
use widgets::{Message, Update, Widget};

enum WidgetAlign {
    Left,
    Right
}

#[derive(PartialEq)]
enum WidgetStatus {
    Okay,
    PendingRender,
    PendingMeasure,
    Measured
}

struct WidgetState {
    widget: Box<Widget>,
    align: WidgetAlign,
    position: u16,
    width: u16,
    status: WidgetStatus
}

impl WidgetState {
    fn new(widget: Box<Widget>, align: WidgetAlign) -> Self {
        WidgetState {
            widget: widget,
            align: align,
            position: 0,
            width: 0,
            status: WidgetStatus::PendingMeasure
        }
    }

    fn should_render(&self) -> bool {
        match self.status {
            WidgetStatus::Measured => true,
            WidgetStatus::PendingRender => true,
            _ => false
        }
    }

    fn update(&mut self, status: WidgetStatus) {
        if self.status == WidgetStatus::PendingMeasure && status == WidgetStatus::PendingRender {
        }
        else {
            self.status = status;
        }
    }

    fn measure(&mut self) -> bool {
        if self.status == WidgetStatus::PendingMeasure {
            self.update(WidgetStatus::Measured);
            let old_width = self.width;
            self.width = self.widget.width();
            if old_width != self.width {
                return true;
            }
        }
        false
    }

    fn contains(&self, x: u16) -> bool {
        x >= self.position && x < self.position + self.width
    }
}

pub struct Panel {
    window: Window,
    conn: x11::Connection,
    context: Context,
    widgets: Vec<WidgetState>,
    finishing: bool
}

impl Panel {
    pub fn new(conn: x11::Connection, cfg: &Config) -> Panel {
        let window = Window::new(conn.clone(), cfg);
        let context = window.make_context();

        Panel {
            window: window,
            conn: conn,
            context: context,
            widgets: vec![],
            finishing: false
        }
    }

    pub fn create(&mut self) {
        self.window.create();
        for state in self.widgets.iter_mut() {
            state.widget.init();
        }
        self.conn.flush();
    }

    pub fn make_context(&self) -> Context {
        self.context.clone()
    }

    pub fn handle_event(&mut self, event: Message) -> bool {
        let finishing = self.finishing;
        let mut should_exit = false;
        let mut should_relayout = false;
        let mut press_location = None;

        if let Message::XcbEvent(ref event) = event {
            if finishing && self.window.handle_event(event) {
                return true
            }
            if event.response_type() == xcb::EXPOSE {
                should_relayout = true;
            }
            if event.response_type() == xcb::BUTTON_PRESS {
                let event: &xcb::ButtonPressEvent = unsafe { xcb::cast_event(event) };
                let x = event.root_x() as u16;
                press_location = Some(x);
            };
        }
        else if let Message::Quit = event {
            self.finish()
        }

        if !finishing {
            let mut delegate_event = |state: &mut WidgetState, event: &Message| {
                match state.widget.handle_event(event) {
                    Update::Quit =>
                        should_exit = true,
                    Update::Relayout =>
                        state.update(WidgetStatus::PendingMeasure),
                    Update::Rerender =>
                        state.update(WidgetStatus::PendingRender),
                    Update::Nothing => ()
                };
            };
            if let Some(x) = press_location {
                for mut state in self.widgets.iter_mut() {
                    if state.contains(x) {
                        let offset_x = x - state.position;
                        let event = Message::MousePress(offset_x);
                        delegate_event(&mut state, &event);
                    }
                }
            }
            for mut state in self.widgets.iter_mut() {
                delegate_event(&mut state, &event);
            }
        }

        // check widgets that requested relayout if widths have changed
        if self.widgets.iter_mut().any(|state| state.measure()) {
            should_relayout = true;
        }

        if should_relayout {
            self.relayout();
        }
        else {
            self.rerender();
        }
        should_exit
    }

    pub fn add_left_widget(&mut self, widget: Box<Widget>) {
        self.widgets.push(WidgetState::new(widget, WidgetAlign::Left));
    }

    pub fn add_right_widget(&mut self, widget: Box<Widget>) {
        self.widgets.insert(0, WidgetState::new(widget, WidgetAlign::Right));
    }

    pub fn relayout(&mut self) {
        self.context.draw_bg(0, self.window.width);

        // measure
        let mut total_width = 0;
        for state in self.widgets.iter_mut() {
            state.measure();
            if !state.widget.fit_width() {
                total_width += state.width;
            }
        }
        let margin = self.window.width.saturating_sub(total_width);
        for state in self.widgets.iter_mut() {
            if state.widget.fit_width() {
                state.width = margin;
            }
        }

        // position
        let mut left_pos = 0;
        let mut right_pos = self.window.width;
        for state in self.widgets.iter_mut() {
            match state.align {
                WidgetAlign::Left => {
                    state.position = left_pos;
                    left_pos += state.width;
                },
                WidgetAlign::Right => {
                    right_pos -= state.width;
                    state.position = right_pos;
                }
            }
        }

        // render
        for state in self.widgets.iter_mut() {
            state.widget.render(state.position, state.width);
            state.update(WidgetStatus::Okay);
        }
        self.conn.flush();
    }

    pub fn rerender(&mut self) {
        for state in self.widgets.iter_mut() {
            if state.should_render() {
                self.context.draw_bg(state.position, state.width);
                state.widget.render(state.position, state.width);
                state.update(WidgetStatus::Okay);
            }
        }
        self.conn.flush();
    }

    pub fn finish(&mut self) {
        self.finishing = true;
        for state in self.widgets.iter_mut() {
            state.widget.finish();
        }
        self.window.finish();
    }
}
