use std::rc::Rc;

use xcb;
use xcb::render::Color;

use config::Config;
use ui::color;
use ui::context::Context;
use ui::ext;
use ui::ext::ConnectionExt;
use ui::x11;
use ui::font;

pub struct Window {
    conn: x11::Connection,
    pub window: xcb::Window,
    pub height: u16,
    pub width: u16,
    pub text_color: Color,
    pub bg_color: Color,
    fonts: Rc<font::FontLoader>,
    picture: xcb::render::Picture
}

impl Window {
    pub fn new(conn: x11::Connection, cfg: &Config) -> Self {
        let conn = conn;
        let window = conn.generate_id();
        let picture = conn.generate_id();

        let height = cfg.lookup("bar.height").and_then(|v| v.as_integer()).expect("bar.height is not set") as u16;
        let text_color = cfg.lookup("bar.text_color")
            .and_then(|v| v.as_str())
            .and_then(|s| color::from_str(s))
            .unwrap_or(color::WHITE);
        let bg_color = cfg.lookup("bar.bg_color")
            .and_then(|v| v.as_str())
            .and_then(|s| color::from_str(s))
            .unwrap_or(color::BLACK);
        let width = conn.default_screen().width;
        let font_loader = Rc::new(font::FontLoader::from_config(conn.clone(), cfg));

        Window {
            conn: conn,
            height: height,
            width: width,
            text_color: text_color,
            bg_color: bg_color,
            window: window,
            fonts: font_loader,
            picture: picture
        }
    }

    pub fn create(&mut self) {
        let (root, width) = {
            let screen = self.conn.default_screen();
            (screen.root, screen.width)
        };

        xcb::create_window(
            &self.conn,
            xcb::COPY_FROM_PARENT as u8,
            self.window,
            root,
            0, 0,
            width, self.height,
            0,
            xcb::WINDOW_CLASS_INPUT_OUTPUT as u16,
            xcb::COPY_FROM_PARENT,
            &[
                (xcb::CW_BACK_PIXEL, color::as_u32(self.bg_color)),
                (xcb::CW_EVENT_MASK, xcb::EVENT_MASK_EXPOSURE | xcb::EVENT_MASK_PROPERTY_CHANGE | xcb::EVENT_MASK_STRUCTURE_NOTIFY | xcb::EVENT_MASK_BUTTON_PRESS)
            ]
        );
        self.set_property(
            self.conn.atom(x11::_NET_WM_WINDOW_TYPE),
            xcb::ATOM_ATOM,
            32,
            &[self.conn.atom(x11::_NET_WM_WINDOW_TYPE_DOCK)]
        );
        xcb::map_window(&self.conn, self.window);
        xcb::render::create_picture(&self.conn, self.picture, self.window, self.conn.get_pict_format(ext::PictFormat::RGB24), &[
            (xcb::render::CP_POLY_EDGE, xcb::render::POLY_EDGE_SMOOTH),
            (xcb::render::CP_POLY_MODE, xcb::render::POLY_MODE_IMPRECISE)
        ]);
        self.conn.flush();
    }

    pub fn make_context(&self) -> Context {
        Context::new(
            self.conn.clone(),
            self.height,
            self.window,
            self.picture,
            self.text_color,
            self.bg_color,
            self.fonts.clone()
        )
    }

    pub fn set_property<T>(&self, name: xcb::Atom, type_: xcb::Atom, format: u8, data: &[T]) {
        xcb::change_property(
            &self.conn,
            xcb::PROP_MODE_REPLACE as u8,
            self.window,
            name,
            type_,
            format,
            data
        );
    }

    pub fn handle_event(&mut self, event: &xcb::GenericEvent) -> bool {
        if event.response_type() == xcb::DESTROY_NOTIFY {
            let event: &xcb::DestroyNotifyEvent = unsafe { xcb::cast_event(event) };
            if event.window() == self.window {
                return true
            }
        }
        false
    }

    pub fn finish(&mut self) {
        xcb::destroy_window(&self.conn, self.window);
        self.conn.flush();
    }
}
