#![allow(dead_code)]

use xcb;
use std::cell::RefCell;
use std::collections::HashMap;
use std::mem;
use std::ops::Deref;
use std::sync::Arc;

macro_rules! atoms {
    ( $( $x:ident ),* ) => {
        #[allow(non_snake_case)]
        $(pub const $x: &'static str = stringify!($x);)*
    }
}

atoms!(
    _NET_ACTIVE_WINDOW,
    _NET_SYSTEM_TRAY_S0,
    _NET_SYSTEM_TRAY_OPCODE,
    _NET_WM_NAME,
    _NET_WM_WINDOW_TYPE,
    _NET_WM_WINDOW_TYPE_DOCK,
    MANAGER,
    UTF8_STRING
);

pub struct WinClass {
    pub instance: String,
    pub class: String
}

#[derive(Clone)]
pub struct Connection {
    conn: Arc<xcb::Connection>,
    default_screen: i32,
    cache: RefCell<HashMap<String, xcb::Atom>>
}

impl Connection {
    pub fn new() -> Option<Connection> {
        if let Ok((conn, default_screen)) = xcb::Connection::connect(None) {
            Some(Connection {
                conn: Arc::new(conn),
                default_screen: default_screen,
                cache: RefCell::new(HashMap::new())
            })
        }
        else {
            None
        }
    }

    pub fn atom(&self, name: &str) -> xcb::Atom {
        let mut cache = self.cache.borrow_mut();
        if cache.contains_key(name) {
            *cache.get(name).unwrap()
        }
        else {
            let atom = xcb::intern_atom(&self.conn, false, name).get_reply().unwrap().atom();
            cache.insert(name.to_string(), atom);
            atom
        }
    }

    pub fn default_screen<'a>(&'a self) -> Screen<'a> {
        self.screen(self.default_screen as usize)
    }

    pub fn screen<'a>(&'a self, index: usize) -> Screen<'a> {
        let setup = self.conn.get_setup();
        let screen = setup.roots().nth(index).unwrap();
        Screen {
            conn: &self,
            root: screen.root(),
            width: screen.width_in_pixels(),
            index: index
        }
    }

    pub fn watch(&self, win: xcb::Window, enabled: bool) {
        let value = if enabled {
            xcb::EVENT_MASK_PROPERTY_CHANGE
        }
        else {
            xcb::EVENT_MASK_NO_EVENT
        };
        xcb::change_window_attributes(&self.conn, win, &[(xcb::CW_EVENT_MASK, value)]);
    }

    pub fn get_window_name(&self, win: xcb::Window) -> String {
        let cookie = xcb::get_property(&self.conn, false, win, self.atom(_NET_WM_NAME), self.atom(UTF8_STRING), 0, 100);
        let reply = cookie.get_reply();
        let value: &str = match reply {
            Ok(reply) => unsafe { mem::transmute(reply.value::<u8>()) },
            _ => ""
        };
        if value == "" {
            let cookie = xcb::get_property(&self.conn, false, win, xcb::ATOM_WM_NAME, xcb::ATOM_STRING, 0, 100);
            let reply = cookie.get_reply();
            let value: &str = match reply {
                Ok(reply) => unsafe { mem::transmute(reply.value::<u8>()) },
                _ => ""
            };
            value.to_string()
        }
        else {
            value.to_string()
        }
    }

    pub fn get_window_class(&self, win: xcb::Window) -> WinClass {
        let cookie = xcb::get_property(&self.conn, false, win, xcb::ATOM_WM_CLASS, xcb::ATOM_STRING, 0, 100);
        let reply = cookie.get_reply();
        let value: &str = match reply {
            Ok(reply) => unsafe { mem::transmute(reply.value::<u8>()) },
            _ => ""
        };
        let parts: Vec<&str> = value.split_terminator('\0').collect();
        if parts.len() == 2 {
            WinClass {
                instance: parts[0].to_string(),
                class: parts[1].to_string()
            }
        }
        else {
            WinClass {
                instance: "".to_string(),
                class: "".to_string()
            }
        }
    }

    pub fn event_loop(&self, handler: &mut FnMut(&xcb::GenericEvent)) {
        loop {
            match self.conn.wait_for_event() {
                None => { break; }
                Some(event) => {
                    handler(&event)
                }
            }
        }
    }
}

impl Deref for Connection {
    type Target = xcb::Connection;

    fn deref(&self) -> &xcb::Connection {
        &self.conn
    }
}

pub struct Screen<'a> {
    conn: &'a Connection,
    pub root: xcb::Window,
    pub width: u16,
    pub index: usize
}

impl<'a> Screen<'a> {
    pub fn get_active_window(&self) -> xcb::Window {
        let cookie = xcb::get_property(self.conn, false, self.root, self.conn.atom(_NET_ACTIVE_WINDOW), xcb::ATOM_WINDOW, 0, 4);
        let reply = cookie.get_reply().unwrap();
        if reply.value_len() == 0 {
            0
        }
        else {
            reply.value::<u32>()[0]
        }
    }

    pub fn search_by_name(&self, name: &str) -> xcb::Window {
        let cookie = xcb::query_tree(&self.conn, self.root);
        let reply = cookie.get_reply().unwrap();
        for w in reply.children() {
            if self.conn.get_window_name(*w) == name {
                return *w;
            }
        }
        return 0;
    }

    pub fn search_by_class(&self, class_name: &str) -> xcb::Window {
        let cookie = xcb::query_tree(&self.conn, self.root);
        let reply = cookie.get_reply().unwrap();
        for w in reply.children() {
            if self.conn.get_window_class(*w).class == class_name {
                return *w;
            }
        }
        return 0;
    }
}
